import itertools


class Register:
    instance_counter = itertools.count()
    prefix = 'r'

    def __init__(self, size, name=None):
        if size <= 0 or not isinstance(size, int):
            raise Exception('Invalid register size "{}"'.format(size))

        self.size = size

        if name is None:
            name = '{}{}'.format(self.prefix, next(self.instance_counter))

        self.name = name

    def __len__(self):
        return self.size

    def __repr__(self):
        return '{}[{}]'.format(self.name, self.size)



import numpy as np

from qsy import gates
from qsyasm.log import print_info

from .backend import Backend


class CHPBackend(Backend):
    '''
    CHPBackend can efficiently simulate stabilizer circuits (circuits consisting
    of only CNOT, H and phase gates). The algorithm used to achieve this is
    described in a paper by Scott Aaronson and Daniel Gottesman found at
    https://arxiv.org/abs/quant-ph/0406196.
    '''

    SUPPORTED_GATES = [gates.CX, gates.H, gates.S, gates.X, gates.Z, gates.CZ]

    def __init__(self, size, name):
        self.name = name
        self.size = size

        # Initialize tableau
        # X generators
        self.x = np.concatenate(
            (np.eye(self.size, dtype=np.int8),
             np.zeros((self.size, self.size), dtype=np.int8),
             # scratch space
             np.zeros((1, self.size), dtype=np.int8)),
            axis=0
        )

        # Z generators
        self.z = np.concatenate(
            (np.zeros((self.size, self.size), dtype=np.int8),
             np.eye(self.size, dtype=np.int8),
             # scratch space
             np.zeros((1, self.size), dtype=np.int8)),
            axis=0
        )

        # Phase (0 for +1, 1 for i, 2 for -1, 3 for -i)
        self.r = np.zeros((2*self.size + 1, 1), dtype=np.int8)

    def apply_gate(self, gate, *params, adjoint=False):
        if gate not in self.SUPPORTED_GATES:
            supported_gate_names = map(lambda g: g.name, self.SUPPORTED_GATES)
            raise Exception(
                'Unsupported gate "{}" for CHP back-end. '.format(gate.name) +
                'Supported gates are {}.'.format(', '.join(supported_gate_names))
            )

        if gate.arity == 1:
            target = params[0]
        else:
            control, target = params

            self._check_in_range(control)

        self._check_in_range(target)

        if gate == gates.CX:
            self._cnot(control, target)
        elif gate == gates.H:
            self._h(target)
        elif gate == gates.S:
            if adjoint:
                # S^dagger = SSS
                self._s(target)
                self._s(target)

            self._s(target)
        elif gate == gates.X:
            self._x(target)
        elif gate == gates.Z:
            self._z(target)
        elif gate == gates.CZ:
            self._cz(control, target)

    def measure(self, target):
        p = None
        for i in range(self.size, 2*self.size):
            if self.x[i, target] == 1:
                p = i
                break

        if p is not None:
            # Measurement outcome is probabilistic
            for i in range(2*self.size):
                if i != p and self.x[i, target] == 1:
                    self._rowsum(i, p)

            # Set (p−n)th row equal to pth row
            self.x[p-self.size] = self.x[p]
            self.z[p-self.size] = self.z[p]
            self.r[p-self.size] = self.r[p]

            self.x[p] = 0
            self.z[p] = 0

            measurement = np.random.randint(2)
            self.r[p] = measurement

            self.z[p, target] = 1

            return measurement
        else:
            # Measurement outcome is deterministic
            self.x[2*self.size] = 0
            self.z[2*self.size] = 0
            self.r[2*self.size] = 0

            for i in range(self.size):
                if self.x[i, target] == 1:
                    self._rowsum(2*self.size, i + self.size)

            return int(self.r[2*self.size])

    def measure_all(self):
        return [self.measure(i) for i in range(self.size)]

    def yield_state(self):
        print_info(
            'Printing quantum state is not supported by the CHP back-end. ' +
            'Use the statevector back-end if you wish to see the quantum state.'
        )
        yield from []

    def to_dirac(self):
        # TODO: make CHP state printable/readable
        return ''

    def _h(self, target):
        for i in range(2*self.size):
            self.r[i] ^= self.x[i, target] & self.z[i, target]
            self.x[i, target], self.z[i, target] = self.z[i, target], self.x[i, target]

    def _cnot(self, control, target):
        for i in range(2*self.size):
            self.r[i] ^= self.x[i, control] & self.z[i, target]
            self.r[i] &= self.x[i, target] ^ self.z[i, control] ^ 1

            self.x[i, target] ^= self.x[i, control]
            self.z[i, control] ^= self.z[i, target]

    def _s(self, target):
        for i in range(2*self.size):
            self.r[i] ^= self.x[i, target] & self.z[i, target]
            self.z[i, target] ^= self.x[i, target]

    def _x(self, target):
        self._h(target)
        self._z(target)
        self._h(target)

    def _z(self, target):
        self._s(target)
        self._s(target)

    def _cz(self, control, target):
        self._h(target)
        self._cnot(control, target)
        self._h(target)

    def _rowsum(self, h, i):
        '''
        Set generator h equal to h + i. Keep track of the factors of i that r_h
        goes through when multiplying Pauli matrices.
        '''
        def g(x1, z1, x2, z2):
            '''
            Return the exponent to which i is raised (either 0, 1, or −1) when
            the Pauli matrices represented by x1 z1 and x2 z2 are multiplied.
            '''
            if x1 == z1 == 0:
                return 0
            if x1 == z1 == 1:
                return z2 - x2
            if x1 == 1 and z1 == 0:
                return z2 * (2*x2 - 1)
            if x1 == 0 and z1 == 1:
                return x2 * (1 - 2*z2)

        rowsum = sum(g(self.x[i, j], self.z[i, j], self.x[h, j], self.z[h, j])
                     for j in range(self.size))
        rowsum += 2*self.r[h] + 2*self.r[i]

        if rowsum == 0:
            self.r[h] = 0
        elif rowsum == 2 % 4:
            self.r[h] = 1

        for j in range(self.size):
            self.x[h, j] ^= self.x[i, j]
            self.z[h, j] ^= self.z[i, j]



import numpy as np


def format_complex(c):
    fmt = '{:+.5f}'.format(c.real).rstrip('0').rstrip('.')

    if not np.isclose(c.imag, 0):
        imag_fmt = '{:+.5f}'.format(c.imag).rstrip('0').rstrip('.') + 'i'
        if np.isclose(c.real, 0.0):
            fmt = imag_fmt
        else:
            fmt += imag_fmt

    return fmt



import sys

NO_COLOR = '\33[m'
BOLD = '\033[1m'
RED, GREEN, ORANGE = ('\33[{}m'.format(i) for i in range(31, 34))


def error_fmt(message):
    return '{}{}error{}: {}'.format(BOLD, RED, NO_COLOR, message)


def warning_fmt(message):
    return '{}{}warning{}: {}'.format(BOLD, ORANGE, NO_COLOR, message)


def info_fmt(message):
    return '{}{}info{}: {}'.format(BOLD, GREEN, NO_COLOR, message)


def print_error(message):
    print(error_fmt(message), file=sys.stderr)


def print_warning(message):
    print(warning_fmt(message), file=sys.stderr)


def print_info(message):
    print(info_fmt(message), file=sys.stderr)



from .error import QsyASMError
from qsy import QuantumRegister, ClassicalRegister


class Env:
    def __init__(self):
        self.qrs = {}
        self.crs = {}

    def qr(self, name):
        if name not in self.qrs:
            raise IndexError('Undefined quantum register "{}"'.format(name))

        return self.qrs[name]

    def cr(self, name):
        if name not in self.crs:
            raise IndexError('Undefined classical register "{}"'.format(name))

        return self.crs[name]

    def create_qr(self, name, size, backend):
        self.qrs[name] = QuantumRegister(size, name, backend=backend)

    def create_cr(self, name, size):
        self.crs[name] = ClassicalRegister(size, name)



class ParseError(Exception):
    def __init__(self, msg, lexpos, lineno):
        super(ParseError, self).__init__(msg)
        self.msg = msg
        self.lexpos = lexpos
        self.lineno = lineno


class QsyASMError(RuntimeError):
    pass



import collections

Gate = collections.namedtuple(
    'Gate', ['name', 'matrix', 'adjoint_matrix', 'arity']
)


def C(gate):
    """Create a controlled-U gate."""
    return Gate('C{}'.format(gate.name), gate.matrix, gate.adjoint_matrix, 2)


def CC(gate):
    """Create a controlled-controlled-U gate."""
    return Gate('CC{}'.format(gate.name), gate.matrix, gate.adjoint_matrix, 3)



import numpy as np
import time
from collections import defaultdict

from qsy import gates, __version__
from qsy.util import format_complex
from qsy.backends import StatevectorBackend, CHPBackend

from .error import ParseError, QsyASMError
from .parser import QsyASMParser
from .instruction import Operation
from .env import Env
from .log import print_warning, print_info

OPERATION_GATES = {
    Operation.I: gates.I,
    Operation.X: gates.X,
    Operation.Y: gates.Y,
    Operation.Z: gates.Z,

    Operation.H: gates.H,
    Operation.S: gates.S,
    Operation.CX: gates.CX,
    Operation.CY: gates.CY,
    Operation.CZ: gates.CZ,

    Operation.T: gates.T,
    Operation.RX: gates.Rx,
    Operation.RY: gates.Ry,
    Operation.RZ: gates.Rz,
    Operation.CRX: gates.CRx,
    Operation.CRY: gates.CRy,
    Operation.CRZ: gates.CRz,
    Operation.CCX: gates.CCX
}


class QsyASMProgram:
    MAX_PRINTABLE_QUBITS = 16

    def __init__(self, args):
        self.filename = args['filename']

        try:
            with open(self.filename) as f:
                self.input = f.read()
        except FileNotFoundError as e:
            raise QsyASMError('Error reading input: {}'.format(str(e)))

        self.time = args['time']
        self.verbose = args['verbose']
        self.ignore_print_warning = args['ignore_print_warning']
        self.skip_zero_amplitudes = args['skip_zero_amplitudes']

        self.shots = args['shots']
        self.measurement_results = {}

        self.backend_arg = args['backend']
        if self.backend_arg == 'chp':
            self.backend = CHPBackend
        else:
            self.backend = StatevectorBackend

        self._verbose_print('Using {} backend'.format(self.backend.__name__))

        self.parser = QsyASMParser()
        self.env = Env()

    def run(self):
        print_info('qsyasm v{}'.format(__version__))
        print_info('A state vector/stabilizer circuit simulator assembly runner')
        print_info('=' * 60)

        start = time.time()

        try:
            instructions = self.parser.parse(self.input)
        except ParseError as e:
            raise QsyASMError(self._error_message(e.msg, e.lexpos, e.lineno))

        # TODO: we can auto-detect stabilizer circuits and use the CHP back-end
        # when it is capable of printing its state so no features are lost.
        #  if self._can_use_chp(instructions) and self.backend_arg is None:
        #      print_info('Stabilizer circuit detected, using CHP back-end')
        #      self.backend = CHPBackend

        self._verbose_print('Executing {} shots'.format(self.shots))

        for _ in range(self.shots):
            self.eval(instructions)
            self._save_measurements()

        end = time.time()

        self.dump_registers()

        if self.time:
            print_info('Program execution took {:.5f} seconds'.format(end - start))

    def eval(self, instructions):
        for instr in instructions:
            try:
                if instr.is_gate():
                    self._eval_gate(instr)
                elif instr.type == Operation.QR or instr.type == Operation.CR:
                    self._eval_register(instr)
                elif instr.type == Operation.MEASURE:
                    self._eval_measure(instr)
                elif instr.type == Operation.ERROR:
                    raise QsyASMError(
                        self._error_message(
                            'Undefined operation "{}"'.format(instr.op[0]),
                            instr.lexpos,
                            instr.lineno
                        )
                    )
            except QsyASMError as e:
                raise e
            except Exception as e:
                raise QsyASMError(self._error_message(e, instr.lexpos, instr.lineno))

    def dump_registers(self):
        for qr_name, qr in self.env.qrs.items():
            if qr.size > self.MAX_PRINTABLE_QUBITS and not self.ignore_print_warning:
                print_warning(
                    'Quantum register {} too large to print ({} qubits)'.format(
                        qr.name, qr.size) +
                    ' (to print the state anyway, pass ' +
                    '--ignore-print-warning as argument)')
                continue

            print('{}[{}]: {}'.format(qr_name, qr.size, qr.to_dirac()))

            for i, amplitude in qr.yield_state():
                if np.isclose(amplitude, 0) and self.skip_zero_amplitudes:
                    continue

                amplitude = format_complex(amplitude)
                print('{:>8} | {:0{size}b}'.format(amplitude, i, size=qr.size))

        for cr_name, cr in self.env.crs.items():
            if self.shots > 1:
                bits = dict(self.measurement_results[cr_name])
            else:
                bits = ''.join(str(bit) for bit in cr.state)

            print('{}[{}]: {}'.format(cr_name, cr.size, bits))

    def _eval_gate(self, instr):
        gate = OPERATION_GATES[instr.type]
        args = instr.args

        # TODO: multi qubit gates across registers
        register = args[0][0]
        targets = [arg[1] for arg in args]

        if callable(gate):
            gate_arg = instr.op[1]
            gate = gate(gate_arg)

        adjoint_message = 'adjoint ' if instr.adjoint else ''
        self._verbose_print('Applying gate {}{} on {}{}'.format(
                            adjoint_message, gate.name, register, targets))

        self.env.qr(register).apply_gate(gate, *targets, adjoint=instr.adjoint)

    def _eval_register(self, instr):
        if len(instr.op) != 2:
            raise QsyASMError(
                self._error_message(
                    'Missing register size in {} definition'.format(instr.op_name),
                    instr.lexpos, instr.lineno
                )
            )

        register_size = instr.op[1]

        if instr.type == Operation.QR:
            for register_name in instr.args:
                self.env.create_qr(register_name, register_size, self.backend)
        elif instr.type == Operation.CR:
            for register_name in instr.args:
                self.env.create_cr(register_name, register_size)

    def _eval_measure(self, instr):
        qtarget = instr.args[0]
        qtarget_name = qtarget[0]

        if len(instr.args) == 2 and len(instr.args[0]) != len(instr.args[1]):
            raise QsyASMError(
                self._error_message(
                    'Mismatched register sizes in measurement',
                    instr.lexpos,
                    instr.lineno
                )
            )

        if len(qtarget) == 1:
            # Measure all
            measured = self.env.qr(qtarget_name).measure_all()
        elif len(qtarget) == 2:
            # Measure single qubit
            qubit = qtarget[1]
            measured = self.env.qr(qtarget_name).measure(qubit)

        if len(instr.args) == 2:
            # Save measurement to classical register
            ctarget = instr.args[1]

            if len(ctarget) == 1:
                if len(qtarget) == 1:
                    # Ensure the classical and quantum registers are of the same size
                    # when measuring all qubits
                    qtarget_size = self.env.qr(qtarget).size
                    ctarget_size = self.env.cr(ctarget).size

                    if qtarget_size != ctarget_size:
                        raise QsyASMError(
                            self._error_message(
                                'Mismatched register sizes in measurement ({}[{}] and {}[{}])'.format(
                                    qtarget_name, qtarget_size, ctarget[0], ctarget_size
                                ),
                                instr.lexpos,
                                instr.lineno
                            )
                        )

                self.env.cr(ctarget).set_state(measured)
            elif len(ctarget) == 2:
                ctarget, register_index = ctarget
                self.env.cr(ctarget)[register_index] = measured

        # Save measurement results when shots > 1
        if self.shots > 1 and ctarget not in self.measurement_results:
            self.measurement_results[ctarget] = defaultdict(int)

    def _save_measurements(self):
        for cr_name in self.measurement_results.keys():
            cr_value = self.env.cr(cr_name).state
            bit_string = ''.join(str(bit) for bit in cr_value)
            self.measurement_results[cr_name][bit_string] += 1

    def _can_use_chp(self, instructions):
        '''
        Determine if a program is a stabilizer circuit and can use the CHP
        back-end.
        '''
        for instr in instructions:
            if instr.is_gate():
                gate = OPERATION_GATES[instr.type]
                if gate not in CHPBackend.SUPPORTED_GATES:
                    return False

        return True

    def _verbose_print(self, msg):
        if self.verbose:
            print_info(msg)

    def _error_message(self, msg, lexpos, lineno):
        column = self._find_column(lexpos)
        return '{}:{}:{}: {}'.format(self.filename, lineno, column, msg)

    def _find_column(self, lexpos):
        line_start = self.input.rfind('\n', 0, lexpos) + 1
        return (lexpos - line_start) + 1



import ply.lex as lex
from .error import ParseError


class QsyASMLexer:
    tokens = (
        'INTEGER',
        'FLOAT',
        'IDENT',
        'COMMA',
        'LBRACKET',
        'RBRACKET',
        'LPAREN',
        'RPAREN',
        'PLUS',
        'MIN',
        'DIV',
        'POW',
        'MUL',
        'NEWLINE',
        'ADJ'
    )

    reserved = {
        'adj': 'ADJ'
    }

    t_COMMA = r','
    t_LBRACKET = r'\['
    t_RBRACKET = r'\]'
    t_LPAREN = r'\('
    t_RPAREN = r'\)'
    t_PLUS = r'\+'
    t_MIN = r'-'
    t_DIV = r'/'
    t_POW = r'\*\*'
    t_MUL = r'\*'

    t_ignore = '\t\r '
    t_ignore_COMMENT = r';.*'

    def __init__(self, **kwargs):
        self.lexer = lex.lex(module=self, **kwargs)

    def t_IDENT(self, t):
        r'[a-zA-Z_][a-zA-Z0-9_]*'
        t.type = self.reserved.get(t.value, 'IDENT')
        return t

    def t_INTEGER(self, t):
        r'\d+'
        t.value = int(t.value)
        return t

    def t_FLOAT(self, t):
        r'\d+\.\d+'
        t.value = float(t.value)
        return t

    def t_NEWLINE(self, t):
        r'\n+'
        t.lexer.lineno += t.value.count('\n')
        return t

    def t_error(self, t):
        raise ParseError('Unknown token "{}"'.format(t.value[0]), t)


lexer = QsyASMLexer()



import argparse
import sys

from qsy import __version__

from .program import QsyASMProgram
from .error import QsyASMError
from .log import print_error


def main():
    argparser = argparse.ArgumentParser(description='qsyasm assembly runner')

    argparser.add_argument('filename', type=str, help='qsyasm file to execute')
    argparser.add_argument('-V', '--version', action='version',
                           version='%(prog)s v' + __version__)
    argparser.add_argument('-v', '--verbose', action='store_true',
                           help='verbose output')
    argparser.add_argument('-t', '--time', action='store_true',
                           help='time program execution')
    argparser.add_argument('-b', '--backend', choices=('chp', 'statevector'),
                           default=None, metavar='B',
                           help='simulator back-end to use: chp or statevector (default: statevector)')
    argparser.add_argument('-s', '--shots', type=int, default=1,
                           help='amount of shots to run')
    argparser.add_argument('--ignore-print-warning', action='store_true',
                           help='ignore register too large to print warning')
    argparser.add_argument('--skip-zero-amplitudes', action='store_true',
                           help='don\'t print states with an amplitude of 0')

    args = vars(argparser.parse_args())

    try:
        p = QsyASMProgram(args)
        p.run()
    except QsyASMError as e:
        print_error(e)
        sys.exit(-1)



import numpy as np
from .gate import Gate

PAULI_I_MATRIX = np.array([[1, 0],
                           [0, 1]])
PAULI_X_MATRIX = np.array([[0, 1],
                           [1, 0]])
PAULI_Y_MATRIX = np.array([[0, -1j],
                           [1j, 0]])
PAULI_Z_MATRIX = np.array([[1,  0],
                           [0, -1]])

I = Gate('I', PAULI_I_MATRIX, PAULI_I_MATRIX, 1)

X = Gate('X', PAULI_X_MATRIX, PAULI_X_MATRIX, 1)

Y = Gate('Y', PAULI_Y_MATRIX, PAULI_Y_MATRIX, 1)

Z = Gate('Z', PAULI_Z_MATRIX, PAULI_Z_MATRIX, 1)
